import asyncio
import socket
from typing import Optional, Tuple

class TCP:
    def __init__(self, ipv6: bool, proxy: Optional[dict]):
        self.ipv6 = ipv6
        self.proxy = proxy
        self.socket: Optional[socket.socket] = None

    async def connect(self, address: Tuple[str, int]):
        raise NotImplementedError

    async def send(self, data: bytes):
        raise NotImplementedError

    async def recv(self) -> Optional[bytes]:
        raise NotImplementedError

    async def close(self):
        if self.socket:
            self.socket.close()
            self.socket = None

class TCPAbridged(TCP):
    async def connect(self, address: Tuple[str, int]):
        self.socket = socket.socket(socket.AF_INET6 if self.ipv6 else socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(10)
        await asyncio.get_event_loop().run_in_executor(None, self.socket.connect, address)
        # Abridged protocol handshake (send 0xEF as per MTProto abridged spec)
        self.socket.send(b"\xEF")

    async def send(self, data: bytes):
        if not self.socket:
            raise ConnectionError("Socket not connected")
        length = (len(data) // 4).to_bytes(1, "little")
        await asyncio.get_event_loop().run_in_executor(None, self.socket.send, length + data)

    async def recv(self) -> Optional[bytes]:
        if not self.socket:
            raise ConnectionError("Socket not connected")
        length_byte = await asyncio.get_event_loop().run_in_executor(None, self.socket.recv, 1)
        if not length_byte:
            return None
        length = int.from_bytes(length_byte, "little") * 4
        data = b""
        while len(data) < length:
            chunk = await asyncio.get_event_loop().run_in_executor(None, self.socket.recv, length - len(data))
            if not chunk:
                raise ConnectionError("Connection closed by server")
            data += chunk
        return data

class Transport:
    def __init__(self):
        pass


from typing import List
import os
from neongram.parser.tl_object import TLConstructor, TLType
from neongram.parser.tl_parser import TLParameter 

class TLParser:
    def __init__(self, schema_path: str):
        self.schema_path = schema_path
        self.constructors = {}
        self.types = {}
        self.parse_schema()

    def parse_schema(self):
        if not os.path.exists(self.schema_path):
            raise FileNotFoundError(f"Schema file not found: {self.schema_path}")
        with open(self.schema_path, "r") as f:
            for line in f:
                line = line.strip()
                if not line or line.startswith("//"):
                    continue
                if "=" in line:
                    name, rest = line.split("=", 1)
                    crc32_part, rest = rest.split(" ", 1)
                    crc32 = int(crc32_part[1:], 16)  # Extract CRC32 
                    type_name = rest.split(";", 0)[0].strip()
                    params = self._parse_params(rest)
                    self.constructors[name] = TLConstructor(
                        name=name,
                        crc32=crc32,
                        params=params,
                        tl_type=TLType(type_name)
                    )

    def _parse_params(self, param_str: str) -> List[TLParameter]:
        params = []
        param_parts = param_str.split()
        for part in param_parts:
            if ":" in part:
                name, type_name = part.split(":")
                is_flag = type_name.startswith("flags.")  # Simplified flag detection
                flag_bit = int(type_name.split(".")[1].split("?")[0]) if is_flag else None
                param_type = type_name.split("?")[1] if is_flag else type_name
                params.append(
                    TLParameter(
                        name=name,
                        type=param_type,
                        is_flag=is_flag,
                        flag_bit=flag_bit
                    )
                )
        return params
    
class TLParameter:
        def __init__(self, name: str, type: str, is_flag: bool = False, flag_bit: int = None):
          self.name = name
          self.type = type
          self.is_flag = is_flag
          self.flag_bit = flag_bit
          


from functools import wraps
from typing import Dict, Any, Callable

def is_flag(func: Callable) -> Callable:
    """Decorator to mark a TL parameter as a flag and handle its conditional presence.

    Args:
        func: The function or method to decorate (e.g., serialize or deserialize).

    Returns:
        Callable: Wrapped function with flag handling.
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        params = kwargs.get('params', [])
        values = kwargs.get('values', {})
        for param in params:
            if 'flag' in param and param.get('flag') is not None:
                flag_value = values.get(param['flag'], 0)
                if not (flag_value & (1 << param.get('bit', 0))):
                    # If the flag bit is not set, exclude this parameter
                    values.pop(param['name'], None)
        return func(self, *args, **kwargs)
    return wrapper


# neongram/client/client.py
from typing import Optional, Dict, Any, Callable, Union
import asyncio
import platform
import os
import logging
from neongram.network.shared import Connection
from neongram.client.session import Session
from neongram.parser.tl_object import TLFunction
from neongram.parser.tl_parser import TLParser
from neongram import __version__

log = logging.getLogger(__name__)

class NeonClient:
    """A client for interacting with the Telegram API."""

    def __init__(
        self,
        name: str,
        api_id: Optional[Union[int, str]] = None,
        api_hash: Optional[str] = None,
        app_version: str = "Neon 1.0.0",
        device_model: str = f"{platform.python_implementation()} {platform.python_version()}",
        system_version: str = f"{platform.system()} {platform.release()}",
        lang_code: str = "en",
        ipv6: bool = False,
        proxy: Optional[Dict[str, Any]] = None,
        test_mode: bool = False,
        bot_token: Optional[str] = None,
        session_string: Optional[str] = None,
        in_memory: bool = False,
        phone_number: Optional[str] = None,
        phone_code: Optional[str] = None,
        password: Optional[str] = None,
        workers: int = min(32, (os.cpu_count() or 1) + 4),
        workdir: str = os.getcwd(),
        plugins: Optional[Dict[str, Any]] = None,
        parse_mode: str = "both",
        no_updates: bool = False,
        takeout: bool = False,
        sleep_threshold: int = 10,
        hide_password: bool = False,
        max_concurrent_transmissions: int = 1,
    ):
        self.name = name
        self.api_id = int(api_id) if api_id else None
        self.api_hash = api_hash
        self.app_version = app_version
        self.device_model = device_model
        self.system_version = system_version
        self.lang_code = lang_code
        self.ipv6 = ipv6
        self.proxy = proxy
        self.test_mode = test_mode
        self.bot_token = bot_token
        self.session = Session(name, session_string, in_memory, workdir)
        self.phone_number = phone_number
        self.phone_code = phone_code
        self.password = password
        self.workers = workers
        self.plugins = plugins or {}
        self.parse_mode = parse_mode
        self.no_updates = no_updates or takeout
        self.takeout = takeout
        self.sleep_threshold = sleep_threshold
        self.hide_password = hide_password
        self.max_concurrent_transmissions = max_concurrent_transmissions
        self.dc_id = 2  # Default DC; can be updated dynamically
        self.connection = Connection(self.dc_id, test_mode, ipv6, proxy)
        parser_path = os.path.join(workdir, "schema", "api.tl")
        self.parser = TLParser(parser_path)
        self.running = False
        self.loop: Optional[asyncio.AbstractEventLoop] = None

    async def start(self) -> None:
        from neongram.client.auth import Auth
        self.loop = asyncio.get_event_loop()
        await self.connection.connect()
        self.running = True  # Set running to True before invoking any methods
        if not self.session.auth_key or not self.session.server_salt:
            from neongram.parser.tl_object import TLFunction
            req_pq = TLFunction("req_pq_multi", 0xBE7E8EF1, [{"name": "nonce", "type": "int128"}], "ResPQ")
            import os
            nonce = int.from_bytes(os.urandom(16), "little")
            req_pq.values["nonce"] = nonce
            await self.invoke(req_pq)  # Now safe to call since self.running is True
            await self.auth.authenticate()
            self.connection.auth_key = self.session.auth_key
            self.connection.server_salt = self.session.server_salt
        print(f"neongram started successfully created by nobitha with the version of: {__version__}")

    async def stop(self) -> None:
        self.running = False
        if self.connection:
            await self.connection.close()
        if not self.session.in_memory:
            self.session.save()

    async def invoke(self, function: TLFunction) -> bytes:
        if not self.running:
            raise ConnectionError("Client not started")
        serialized = function.serialize()
        await self.connection.send(serialized)
        return await self.connection.receive()

    async def get_me(self) -> Dict[str, Any]:
        from neongram.methods import UsersGetUsers
        request = UsersGetUsers()
        request.values["id"] = [{"_": "inputUserSelf"}]
        response = await self.invoke(request)
        return {"raw": response.hex()}  # Placeholder; add deserialization

    def run(self, coro: Callable) -> None:
        if not self.loop:
            self.loop = asyncio.get_event_loop()
        self.loop.run_until_complete(coro)


import logging
import socket
import asyncio
import ssl
from typing import Optional
from neongram.network.transport import TCPAbridged
from neongram.network.data_center import DataCenter
from neongram.utils.binary_writer import BinaryWriter

log = logging.getLogger(__name__)

class Connection:
    MAX_CONNECTION_ATTEMPTS = 3

    def __init__(self, dc_id: int, test_mode: bool, ipv6: bool, proxy: Optional[dict] = None, media: bool = False):
        self.dc_id = dc_id
        self.test_mode = test_mode
        self.ipv6 = ipv6
        self.proxy = proxy
        self.media = media
        self.host, self.port = DataCenter.get_address(dc_id, test_mode, ipv6, media)
        self.protocol: Optional[TCPAbridged] = None
        self.auth_key: Optional[bytes] = None
        self.server_salt: Optional[int] = None

    async def connect(self):
        for i in range(Connection.MAX_CONNECTION_ATTEMPTS):
            self.protocol = TCPAbridged(self.ipv6, self.proxy)

            try:
                log.info(f"Connecting to {self.host}:{self.port}...")
                await self.protocol.connect((self.host, self.port))
            except OSError as e:
                log.warning(f"Unable to connect due to network issues: {e}")
                await self.protocol.close()
                await asyncio.sleep(1)
            else:
                log.info(f"Connected to {'Test' if self.test_mode else 'Production'} DC{self.dc_id}"
                         f"{' (media)' if self.media else ''} - IPv{'6' if self.ipv6 else '4'}")
                break
        else:
            log.warning("Connection failed! Max retries reached.")
            raise ConnectionError("Failed to establish connection after max retries")

    async def send(self, data: bytes):
        if not self.protocol:
            raise ConnectionError("No active connection")
        writer = BinaryWriter()
        writer.write_int(0xeeeeeeee)  # MTProto header
        writer.write_int(len(data))   # Length of payload
        writer.write_bytes(data)      # Payload
        await self.protocol.send(writer.get_bytes())

    async def receive(self) -> Optional[bytes]:
        if not self.protocol:
            raise ConnectionError("No active connection")
        return await self.protocol.recv()

    async def close(self):
        if self.protocol:
            await self.protocol.close()
            log.info("Disconnected")